﻿using System;
using System.Collections.Generic;
using System.Globalization;

using Platform.Net;

namespace Platform.Geo
{
    // [XmlInclude(typeof(GeoPosition))]
    // [XmlInclude(typeof(GeoPolyline))]
    // [XmlInclude(typeof(GeoPolygon))]
    // [XmlInclude(typeof(GeoCollection))]
    public abstract class GeoShape : GeoBase, IGeoShape
    {
        #region Conversion Helpers

        /// <summary>
        /// Read WKT value
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static double ReadWktValueData(NetStringReader reader)
        {
            // skip white space
            reader.Skip();

            // read token
            var value = reader.ReadToken();
            if (string.IsNullOrEmpty(value))
                throw new FormatException("Missing value in WKT");
            return double.Parse(value, CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Read WKT value
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static double? ReadOptionalWktValueData(NetStringReader reader)
        {
            // skip white space
            reader.Skip();

            // check for double value
            if (!char.IsDigit((char)reader.Peek()) && reader.Peek() != '.')
                return null;

            // read token
            var value = reader.ReadToken();
            if (string.IsNullOrEmpty(value))
                throw new FormatException("Missing value in WKT");
            return double.Parse(value, CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// To Wkt Value Data
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static string ToWktValueData(double value)
        {
            return value.ToString(CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Read WKT Point
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static GeoPosition ReadWktPointData(NetStringReader reader)
        {
            var longitude = ReadWktValueData(reader);
            var latitude = ReadWktValueData(reader);
            var altitude = ReadOptionalWktValueData(reader);
            return new GeoPosition { Latitude = latitude, Longitude = longitude, Altitude = altitude };
        }

        /// <summary>
        /// To Wkt Value Data
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static string ToWktPointData(GeoPosition position)
        {
            if( position.Altitude != null )
                return string.Format("{0} {1} {2}",
                    ToWktValueData(position.Longitude),
                    ToWktValueData(position.Latitude),
                    ToWktValueData(position.Altitude.Value));

            return string.Format("{0} {1}",
                ToWktValueData(position.Longitude),
                ToWktValueData(position.Latitude));
        }

        /// <summary>
        /// Read WKT Line
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static GeoPosition[] ReadWktLineData(NetStringReader reader)
        {
            reader.Skip();
            if (!reader.Read('(')) throw new FormatException("Missing Literal '(' in wkt ");
            reader.Skip();

            var positions = new List<GeoPosition>();
            while (true)
            {
                positions.Add(ReadWktPointData(reader));
                reader.Skip();
                if (!reader.Read(',')) break;
            }

            reader.Skip();
            if (!reader.Read(')')) throw new FormatException("Missing Literal ')' in wkt ");
            reader.Skip();

            return positions.ToArray();
        }

        /// <summary>
        /// Read WKT polygon
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static GeoPosition[][] ReadWktMultiLineData(NetStringReader reader)
        {
            reader.Skip();
            if (!reader.Read('(')) throw new FormatException("Missing Literal '(' in wkt ");
            reader.Skip();

            var positions = new List<GeoPosition[]>();
            while (true)
            {
                positions.Add(ReadWktLineData(reader));
                reader.Skip();
                if (!reader.Read(',')) break;
            }

            reader.Skip();
            if (!reader.Read(')')) throw new FormatException("Missing Literal ')' in wkt ");
            reader.Skip();

            return positions.ToArray();
        }

        /// <summary>
        /// Read WKT multi polygon
        /// </summary>
        /// <param name="reader"></param>
        /// <returns></returns>
        public static GeoPosition[][][] ReadWktMultiPolygonData(NetStringReader reader)
        {
            reader.Skip();
            if (!reader.Read('(')) throw new FormatException("Missing Literal '(' in wkt ");
            reader.Skip();

            var positions = new List<GeoPosition[][]>();
            while (true)
            {
                positions.Add(ReadWktMultiLineData(reader));
                reader.Skip();
                if (!reader.Read(',')) break;
            }

            reader.Skip();
            if (!reader.Read(')')) throw new FormatException("Missing Literal ')' in wkt ");
            reader.Skip();

            return positions.ToArray();
        }

        #endregion

        /// <summary>
        /// Convert to WKT
        /// </summary>
        /// <returns></returns>
        public abstract string ToWKT();

        /// <summary>
        /// Get Bounds
        /// </summary>
        public abstract GeoRectangle GetBounds();
    }
}
